#include "stdafx.h"
#include "timer.h"

//
// File: Rain.cpp
// Rain v.2930 (first release)
//
// VNAP cgeffect2 plugin - made in RVPF[Ricanet VNAP Plugin Factory], Rica
//
// original work : VNAPȸ, AMIHK
/////////////////////////////////////////////////////////////////////////////

#include <ddraw.h>
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <functional>
#include <cassert>
#include <cmath>

using namespace std;

#define CGEFFECT_PITCH 20

// ȭȿ   
static DDSURFACEDESC2       ddsd;

static LPDIRECTDRAWSURFACE7 lpddBack;
static LPDIRECTDRAWSURFACE7 lpddCG;

static LPDWORD pBackSurface;
static LPDWORD pCGSurface;

static LPWORD pwBackSurface;
static LPWORD pwCGSurface;

static DWORD dwSurfacePitch;
static WORD wSurfacePitch;
static int iWindowWidth, iWindowHeight;
static bool bPixelFormatIs16;

#include "Lib.h"

class Rain
{
public:
	float x;
	float y;
	float vy;
	int   size;

	void Reset()
	{
		y    = -40;
		vy   = ((float)rand() / (float)RAND_MAX * (vyMax-vyMin)) + vyMin;
		x    =  (float)rand() / (float)RAND_MAX * xMax;
		size = int(sizeMax * (vy / vyMax));
	}

	void FirstReset()
	{
		Reset();
		y    = (float)rand() / (float)RAND_MAX * yMax*2;
	}

	void Update( float dt )
	{
		//  Ǵ 
		y += vy * dt;
		if( y > yMax*2 )
		{
			Reset();
		}
	}

	float GetOpacity() const
	{
		return size / sizeMax;
	}

public:
	static void InitGlobalValues( int iMaxX, int iMaxY, int iMaxVy, int iMinVy, int iRainSize )
	{
		Rain::xMax    = (float)iMaxX;
		Rain::yMax    = (float)iMaxY;
		Rain::vyMax   = (float)iMaxVy;
		Rain::vyMin   = (float)iMinVy;
		Rain::sizeMax = (float)iRainSize;
	}

protected:
	static float xMax;
	static float yMax;
	static float vyMax;
	static float vyMin;
	static float sizeMax;
};

template< typename SetPixel >
struct RainRenderer
{
	SetPixel setPixel;
	int r, g, b;
	
	RainRenderer()
	{
		r = setPixel.MakeR(192);
		g = setPixel.MakeR(192);
		b = setPixel.MakeR(202);
	}

	void Render( const Rain& rain, float maxAlpha )
	{
		int x = (int)rain.x;
		float maxDistance = rain.size*0.7f;
		float rainOpacity = rain.GetOpacity();

		for( int i=0; i<rain.size; i++ )
		{
			int y = (int)rain.y - i;
			if( 0<=x && x<(int)ddsd.dwWidth-1 && 0<=y && y<(int)ddsd.dwHeight )
			{
				float distance = fabs(i - rain.size*0.3f);
				float opacity  = maxAlpha*rainOpacity * (1 - distance/maxDistance);
				setPixel(x, y, r, g, b, int(opacity*255));
			}
		}
	}

	//ddsd.dwWidth
	//ddsd.dwHeight
};

float Rain::xMax;
float Rain::yMax;
float Rain::vyMax;
float Rain::vyMin;
float Rain::sizeMax;

vector<Rain> rains;

float maxAlpha;

// ݵ Լ ̿ ϰ  ֽñ ٶϴ.
// Լ̳ ڰ ٸ ,   ʽϴ.

//-----------------------------------------------------------------------------
// Name: BeginCGEffect2
// Desc:  ȭȿ ۵  ȣǴ ԼԴϴ.
//		  Լ ǥ Ķ͸ ϰ ʱصξٰ 
//		 UpdateFrame Լ մϴ.
//-----------------------------------------------------------------------------
// strParam				: cgeffect2 Ŀǵ ڿ Է¹ ĶͰ Ʈ ɴϴ.
// lpBackSurface		: ȭ鿡 µǱ  ̹ Ǵ ǥ Դϴ. 
//						   ϴ  BackSurface  ScreenSurface ˴ϴ.
// lpCGSurface			:  ̹ Ǿ ִ ǥԴϴ. ȭȿ ۵Ǵ  â 
//						  ̹ CGSurface Ǿ ֽϴ.
// lpBufferSurface		:   ǥ Ǿ ֽϴ.
extern "C" __declspec(dllexport) int BeginCGEffect2(char* strParam,
													LPDIRECTDRAWSURFACE7 lpBackSurface, 
													LPDIRECTDRAWSURFACE7 lpCGSurface, 
													LPDIRECTDRAWSURFACE7 lpBufferSurface)			
{
	// Lock޼ҵ带  DDSURFACEDESC2 ü ǥ  ؿ  ֽϴ.
	// ̹Ϳ  Ҷ DDESURFACEDESC2 ü   lpSurface մϴ.
	// Lock  Ŀ Unlock޼ҵ ݵ Lock Ǯ մϴ.

	// ǥ  DDSURFACEDESC2 ü մϴ.
	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(ddsd);

	// BackSurface ̹ Ϳ Pitch մϴ.
	lpddBack = lpBackSurface;
	lpddBack->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	pBackSurface = (LPDWORD)ddsd.lpSurface;
	dwSurfacePitch = (DWORD)ddsd.lPitch/4;
	lpddBack->Unlock(NULL);

	iWindowWidth = ddsd.dwWidth;
	iWindowHeight = ddsd.dwHeight;

	// CGSurface ̹ ͸ մϴ.
	lpddCG = lpCGSurface;
	lpddCG->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	pCGSurface = (LPDWORD)ddsd.lpSurface;
	lpddCG->Unlock(NULL);
	
	bPixelFormatIs16 = (ddsd.ddpfPixelFormat.dwRGBBitCount == 16);

	// 16bit 带   
	pwBackSurface = (LPWORD)pBackSurface;
	pwCGSurface = (LPWORD)pCGSurface;
	wSurfacePitch = (WORD)ddsd.lPitch/2;
	
	// !
	srand(GetTickCount());

	// Ķ ȯ
	int nRains = 100;
	int iMaxVy = 30;
	int iMinVy = 5;
	int iRainSize = 300;
	sscanf(strParam, "%d %d %d %d", &nRains, &iMaxVy, &iMinVy, &iRainSize);

	//  ġ ǥ ʱȭ
	Rain::InitGlobalValues(ddsd.dwWidth, ddsd.dwHeight, iMaxVy, iMinVy, iRainSize);
	rains.resize(nRains);
	for_each(rains.begin(), rains.end(), mem_fun_ref(Rain::FirstReset));

	maxAlpha = 0;

	return 0;
}



//-----------------------------------------------------------------------------
// Name: UpdateFrame
// Desc:   Ҷ  Լ ȣǰ  Ŀ â BackSurface
//		   ȭ鿡 µ˴ϴ.
//-----------------------------------------------------------------------------
// bSurfaceRestored		: ȭȿ   ڰ Ǯũ/带
//						  ȯ , ǥ ޸𸮿 ٰ ٽ Ǿ ޸𸮻
//						  ּҰ ٲϴ.   bSurfaceRestored true ,
//						  lpBackSurface  ּҰ ɴϴ.
//						  bSurfaceRestored  false϶ lpBackSurface NULL
//						  ֽϴ.
extern "C" __declspec(dllexport) int UpdateFrame(BOOL bSurfaceRestored,
												 LPDIRECTDRAWSURFACE7 lpBackSurface, 
												 LPDIRECTDRAWSURFACE7 lpCGSurface, 
												 LPDIRECTDRAWSURFACE7 lpBufferSurface)
{
	// ǥġ 缳
	if(bSurfaceRestored)
	{
		ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
		ddsd.dwSize = sizeof(ddsd);

		lpddBack = lpBackSurface;
		lpddBack->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		pBackSurface = (LPDWORD)ddsd.lpSurface;
		dwSurfacePitch = (DWORD)ddsd.lPitch/4;
		lpddBack->Unlock(NULL);

		lpddCG = lpCGSurface;
		lpddCG->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		pCGSurface = (LPDWORD)ddsd.lpSurface;
		lpddCG->Unlock(NULL);
	
		bPixelFormatIs16 = (ddsd.ddpfPixelFormat.dwRGBBitCount == 16);

		pwBackSurface = (LPWORD)pBackSurface;
		pwCGSurface = (LPWORD)pCGSurface;
		wSurfacePitch = (WORD)ddsd.lPitch/2;
	}

	static Timer timer;
	timer.Update();
	float dt = timer.FrameTime();

	maxAlpha += dt/30;
	if( maxAlpha > 0.9f )
	{
		maxAlpha = 0.9f;
	}
	
    for_each(
		rains.begin(),
		rains.end(),
		bind2nd( mem_fun_ref(&Rain::Update), dt)
		);
                      
	// Ϲ ȭȿ     ̹ ϴ κ
	lpddBack->Blt(NULL, lpddCG, NULL, DDBLT_WAIT, NULL);

	if( bPixelFormatIs16 ) // 16bit
	{
		RainRenderer<SetPixel16> renderer;
		for( unsigned i=0; i<rains.size(); i++ )
		{
			renderer.Render(rains[i], maxAlpha);
		}
	}
	else
	{
		RainRenderer<SetPixel32> renderer;
		for( unsigned i=0; i<rains.size(); i++ )
		{
			renderer.Render(rains[i], maxAlpha);
		}
	}

	return 0;
}

//-----------------------------------------------------------------------------
// Name: EndCGEffect2
// Desc: ȭȿ ÿ ȣǴ ԼԴϴ.
//-----------------------------------------------------------------------------
extern "C" __declspec(dllexport) int EndCGEffect2()
{
	lpddBack->Blt(NULL, lpddCG, NULL, DDBLT_WAIT, NULL);
	return 0;
}